from typing import List, Optional
from pydantic import BaseModel

from ...schemas import Access, Error, AccessScope
from ...exceptions import AccessError


def check_field_access(input: BaseModel, access: Access):
    """
    Check access to fields.

    To define scopes of a field, add a list of scopes to the Field defenition in the kwarg scopes.

    Example:
    ```python
    from pydantic import BaseModel, Field

    class AddressRequest(BaseModel):
        name: str = Field(scopes=['elysium.addresses.update.any',])
    ```
    """
    def check(model: BaseModel, input: dict, access: Access, loc: Optional[List[str]] = None):
        if not loc:
            loc = ['body',]

        for key, value in input.items():
            if isinstance(value, dict):
                try:
                    check(getattr(model, key), value, access, loc=loc + [key])

                except AttributeError:
                    pass

            elif key in model.__fields__:
                scopes = [AccessScope.from_str(audience) for audience in model.__fields__[key].field_info.extra.get('scopes', [])]
                change_scopes = [str(scope) for scope in scopes if scope.action != 'read']
                if change_scopes:
                    if not access.token.has_audience(change_scopes):
                        raise AccessError(detail=Error(
                            type='FieldAccessError',
                            code='access_error.field',
                            detail={
                                'loc': loc + [key],
                            },
                        ))

                elif model.__fields__[key].field_info.extra.get('is_critical'):
                    if not access.token.crt:
                        raise AccessError(detail=Error(
                            type='FieldAccessError',
                            code='access_error.field_is_critical',
                            detail={
                                'loc': loc + [key],
                            },
                        ))

    check(input, input.dict(exclude_unset=True), access)



from .models import BaseModel, ModelKind
from .fields import Field
from .utils.pydantic_django import transfer_from_orm, transfer_to_orm
from .utils.pydantic import Reference
from pydantic.fields import Undefined



from typing import Any, Optional
from pydantic import BaseModel


class Error(BaseModel):
    type: Optional[str] = None
    message: Optional[str] = None
    code: Optional[str] = None
    event_id: Optional[str] = None
    detail: Optional[Any] = None



from decimal import Decimal
from pydantic import BaseModel, validator


TWO_PLACES = Decimal(10) ** -2


class AmountPrecision(BaseModel):
    gross: Decimal
    net: Decimal

    def __getitem__(self, name):
        return getattr(self, name)


class Amount(AmountPrecision):
    @validator('gross', 'net')
    def _round_amount(cls, value: Decimal):
        return value.quantize(TWO_PLACES)



from .checks import check_field_access
from .django_to_pydantic import transfer_from_orm, transfer_current_obj as transfer_from_orm_current_obj
from .pydantic_to_django import transfer_to_orm, TransferAction, update_orm
from .pydantic import DjangoORMBaseModel, validate_object, orm_object_validator
from .utils import dict_resolve_obj_to_id



from typing import Any, Optional, List, Set, Type
from datetime import datetime
from pydantic import BaseModel, Field


class AccessScopes(set):
    pass


class Access(BaseModel):
    token: 'AccessToken'
    scope: Optional['AccessScope'] = None
    scopes: Optional[Set['AccessScope']] = None
    sources: List[str] = Field(default=[])
    eids: List[str] = Field(default=[])
    extra: Any = None

    user: Optional[Any] = None

    @property
    def user_id(self) -> str:
        return self.token.sub

    @property
    def jti(self) -> str:
        return self.token.jti

    @property
    def tenant_id(self) -> str:
        return self.token.ten


class AccessToken(BaseModel):
    iss: str = Field(title='Issuer')
    iat: datetime = Field(title='Issued At')
    nbf: datetime = Field(title='Not Before')
    exp: datetime = Field(title='Expire At')
    sub: str = Field(title='Subject (User)')
    ten: str = Field(title='Tenant')
    aud: List[str] = Field(default=[], title='Audiences')
    rls: List[str] = Field(default=[], title='Roles')
    jti: str = Field(title='JWT ID')
    crt: bool = Field(False, title='Critical')

    def has_audience(self, audiences: List[str]) -> Optional[str]:
        for audience in audiences:
            if audience in self.aud:
                return audience

        return

    def has_audiences(self, audiences: List[str]) -> List[str]:
        return [audience for audience in audiences if audience in self.aud]

    def get_scopes(self):
        audiences = [AccessScope.from_str(audience) for audience in self.aud]
        return audiences

    def __contains__(self, item):
        if isinstance(item, AccessScope):
            return str(item) in self.aud

        if isinstance(item, str):
            return item in self.aud

        raise NotImplementedError


class AccessScope(BaseModel):
    service: str
    resource: str
    action: Optional[str] = None
    selector: Optional[str] = None

    @classmethod
    def from_str(cls, scope: str):
        scopes = scope.split('.') + [None, None]
        return cls(
            service=scopes[0],
            resource=scopes[1],
            action=scopes[2],
            selector=scopes[3],
        )

    def __str__(self):
        return '.'.join(
            filter(
                lambda s: s,
                [
                    self.service,
                    self.resource,
                    self.action,
                    self.selector,
                ],
            )
        )

    def __hash__(self) -> int:
        return hash(self.__str__())


Access.update_forward_refs()
AccessToken.update_forward_refs()



from typing import List


def remove_none(d):
    if isinstance(d, dict):
        return {key: remove_none(value) for key, value in d.items() if value is not None}

    else:
        return d


def key_in_dict(keys: List[str], d: dict) -> bool:
    for key in keys:
        if key in d:
            return True

    return False



import warnings
from collections import defaultdict
from enum import Enum
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Type

from async_tools import is_async, sync_to_async
from django.db import models
from django.db.models.fields import Field as DjangoField
from django.db.models.fields.related_descriptors import ManyToManyDescriptor, ReverseManyToOneDescriptor
from django.db.transaction import atomic
from pydantic import BaseModel, Field, SecretStr, validate_model
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, Undefined, UndefinedType
from sentry_tools.decorators import instrument_span
from sentry_tools.span import set_data, set_tag

from ...schemas import Access
from ..pydantic import Reference, get_orm_field_attr, is_orm_field_set
from .checks import check_field_access
from .pydantic import get_sync_matching_filter

try:
    from dirtyfields import DirtyFieldsMixin

except ImportError:

    class DirtyFieldsMixin:
        pass


try:
    from fastapi.exceptions import RequestValidationError

except ImportError:
    from pydantic import ValidationError as RequestValidationError


class TransferAction(Enum):
    CREATE = 'CREATE'
    SYNC = 'SYNC'
    NO_SUBOBJECTS = 'NO_SUBOBJECTS'


def get_subobj_many_to_many(
    val: BaseModel,
    action: TransferAction,
    related_model: Type[models.Model],
    relatedmanager: models.Manager,
    force_create: bool = False,
    allow_creation: bool = True,
):
    q_filter = models.Q()
    if getattr(val, 'id', None):
        q_filter &= models.Q(**{'pk': val.id})

    elif matching := get_sync_matching_filter(val, related_model):
        q_filter &= matching

    else:
        raise NotImplementedError

    try:
        return related_model.objects.get(q_filter)

    except related_model.DoesNotExist:
        if not allow_creation:
            raise

        return related_model()


def get_subobj_many_to_many_with_intermediate(
    val: BaseModel,
    obj_fields: Dict[str, models.Model],
    action: TransferAction,
    related_model: Type[models.Model],
    relatedmanager: models.Manager,
    force_create: bool = False,
):
    obj_manytomany_fields = {**obj_fields}
    if getattr(val, 'id', None):
        obj_manytomany_fields[relatedmanager.target_field.attname] = val.id

    else:
        raise NotImplementedError

    if force_create or action == TransferAction.CREATE:
        return related_model(**obj_manytomany_fields)

    elif action == TransferAction.SYNC:
        try:
            return related_model.objects.get(**obj_manytomany_fields)

        except related_model.DoesNotExist:
            return get_subobj_many_to_many_with_intermediate(
                val,
                obj_fields,
                action,
                related_model,
                relatedmanager,
                force_create=True,
            )

    else:
        raise NotImplementedError


def get_subobj_rev_many_to_one(
    val: BaseModel,
    obj_fields: Dict[str, models.Model],
    action: TransferAction,
    related_model: Type[models.Model],
    field: Field,
    force_create: bool = False,
):
    q_filter = models.Q()
    try:
        if getattr(val, 'id', None):
            q_filter &= models.Q(**{'pk': val.id})

        elif matching := get_sync_matching_filter(val, related_model, field, obj_fields):
            q_filter &= matching

    except ValueError:
        # its okay to have no_fields_for_matching_defined
        pass

    if action == TransferAction.SYNC and not getattr(val, 'id', None) and not q_filter:
        force_create = True

    if force_create or action == TransferAction.CREATE:
        create_fields = {}
        if getattr(val, 'id', None):
            create_fields['id'] = getattr(val, 'id', None)

        return related_model(**obj_fields, **create_fields)

    elif action == TransferAction.SYNC:
        try:
            if getattr(val, 'id', None):
                return related_model.objects.get(id=val.id, **obj_fields)

            elif q_filter:
                return related_model.objects.filter(**obj_fields).get(q_filter)

            else:
                raise NotImplementedError

        except related_model.DoesNotExist:
            return get_subobj_rev_many_to_one(val, obj_fields, action, related_model, field, force_create=True)

    else:
        raise NotImplementedError


@instrument_span(
    op='transfer_to_orm',
    description=lambda pydantic_obj, django_obj, *args, **kwargs: f'{pydantic_obj} to {django_obj}',
)
def transfer_to_orm(
    pydantic_obj: BaseModel,
    django_obj: models.Model,
    *,
    action: Optional[TransferAction] = None,
    exclude_unset: bool = False,
    access: Optional[Access] = None,
    created_submodels: Optional[List[models.Model]] = None,
    _just_return_objs: bool = False,
    do_not_save_if_no_change: bool = False,
) -> Optional[Tuple[List[models.Model], List[models.Model]]]:
    """
    Transfers the field contents of pydantic_obj to django_obj.
    For this to work it is required to have orm_field set on all of the pydantic_obj's fields, which has to point to the django model attribute.

    It also works for nested pydantic models which point to a field on the **same** django model.

    Example:

    ```python
    from pydantic import BaseModel, Field
    from django.db import models

    class Address(models.Model):
        name = models.CharField(max_length=56)

    class AddressRequest(BaseModel):
        name: str = Field(orm_field=Address.name)
    ```
    """
    if is_async():
        return sync_to_async(transfer_to_orm)(
            pydantic_obj=pydantic_obj,
            django_obj=django_obj,
            action=action,
            exclude_unset=exclude_unset,
            access=access,
            created_submodels=created_submodels,
        )

    set_tag('transfer_to_orm.action', action)
    set_tag('transfer_to_orm.exclude_unset', exclude_unset)
    set_data('transfer_to_orm.access', access)
    set_data('transfer_to_orm.pydantic_obj', pydantic_obj)
    set_data('transfer_to_orm.django_obj', django_obj)

    if created_submodels:
        warnings.warn("Use transfer_to_orm with kwarg action instead of created_submodels", category=DeprecationWarning)
        action = TransferAction.CREATE

    if not action:
        warnings.warn("Use transfer_to_orm with kwarg action", category=DeprecationWarning)

    subobjects: List[models.Model] = created_submodels or []
    objects_to_delete = []
    many_to_many_objs: Dict[models.Manager, List[models.Model]] = {}

    if access:
        check_field_access(pydantic_obj, access)

    pydantic_values: Optional[dict] = pydantic_obj.dict(exclude_unset=True) if exclude_unset else None

    def populate_default(pydantic_cls: BaseModel, django_obj):
        for key, field in pydantic_cls.__fields__.items():
            orm_field = get_orm_field_attr(field.field_info, 'orm_field')
            if not orm_field and issubclass(field.type_, BaseModel):
                populate_default(field.type_, django_obj)

            else:
                if not is_orm_field_set(field.field_info):
                    continue

                if get_orm_field_attr(field.field_info, 'orm_method'):
                    # Do not raise error when orm_method is set
                    continue

                orm_field: DjangoField = get_orm_field_attr(field.field_info, 'orm_field')

                assert orm_field, "orm_field not set on %r of %r" % (field, pydantic_cls)

                setattr(
                    django_obj,
                    orm_field.field.attname,
                    (
                        field.field_info.default
                        if field.field_info.default is not Undefined and field.field_info.default is not ...
                        else None
                    ),
                )

    fields = pydantic_obj.__fields__.items()
    if exclude_unset:
        # XXX: is filtering the fields at this point correct? Tests required
        fields = [(key, field) for key, field in fields if key in pydantic_values]

    for key, field in fields:
        orm_field = get_orm_field_attr(field.field_info, 'orm_field')
        orm_method = get_orm_field_attr(field.field_info, 'orm_method')

        if (
            key == 'id'
            and ((not orm_field or isinstance(orm_field, UndefinedType)) or orm_field.field.attname == key)
            and not orm_method
        ):
            continue

        if orm_method:
            if exclude_unset and key not in pydantic_values:
                continue

            value = getattr(pydantic_obj, field.name)
            if isinstance(value, SecretStr):
                value = value.get_secret_value()

            orm_method(django_obj, value)
            continue

        if not is_orm_field_set(field.field_info):
            continue

        if not orm_field and not (field.shape == SHAPE_SINGLETON and issubclass(field.type_, BaseModel)):
            raise AttributeError("orm_field not found on %r" % field)

        value = getattr(pydantic_obj, field.name)
        if field.shape == SHAPE_SINGLETON:
            if not orm_field and issubclass(field.type_, BaseModel):
                if value is None:
                    if exclude_unset and key not in pydantic_values:
                        continue

                    populate_default(field.type_, django_obj)

                elif isinstance(value, BaseModel):
                    sub_transfer = transfer_to_orm(
                        pydantic_obj=value,
                        django_obj=django_obj,
                        exclude_unset=exclude_unset,
                        access=access,
                        action=action,
                        _just_return_objs=True,
                    )
                    subobjects += sub_transfer[0]
                    objects_to_delete += sub_transfer[1]

                else:
                    raise NotImplementedError

            else:
                if exclude_unset and key not in pydantic_values:
                    continue

                if orm_field.field.is_relation and isinstance(value, models.Model):
                    value = value.pk

                if isinstance(orm_field.field, models.JSONField) and value:
                    if isinstance(value, BaseModel):
                        value = value.dict()

                    elif isinstance(value, dict):
                        pass

                    else:
                        raise NotImplementedError

                setattr(django_obj, orm_field.field.attname, value)

        elif field.shape == SHAPE_LIST:
            if value is None:
                continue

            is_direct_m2m = False
            related_model: Type[models.Model]
            relatedmanager: models.Manager

            if isinstance(orm_field, ManyToManyDescriptor):
                relatedmanager = getattr(django_obj, orm_field.field.attname)

                if hasattr(relatedmanager, 'through') and relatedmanager.through._meta.auto_created:
                    obj_fields = relatedmanager.core_filters
                    related_model = relatedmanager.model
                    get_subobj = partial(
                        get_subobj_many_to_many,
                        action=action,
                        related_model=related_model,
                        relatedmanager=relatedmanager,
                        allow_creation=action in (TransferAction.CREATE, TransferAction.SYNC),
                    )
                    is_direct_m2m = True

                else:
                    obj_fields = {relatedmanager.source_field_name: django_obj}
                    related_model = relatedmanager.through
                    get_subobj = partial(
                        get_subobj_many_to_many_with_intermediate,
                        obj_fields=obj_fields,
                        action=action,
                        related_model=related_model,
                        relatedmanager=relatedmanager,
                    )

            elif isinstance(orm_field, ReverseManyToOneDescriptor):
                relatedmanager = getattr(django_obj, orm_field.rel.name)
                related_model = relatedmanager.field.model
                obj_fields = {relatedmanager.field.name: django_obj}

                get_subobj = partial(
                    get_subobj_rev_many_to_one,
                    obj_fields=obj_fields,
                    action=action,
                    related_model=related_model,
                    field=field,
                )

            else:
                raise NotImplementedError

            existing_object_ids = set(related_model.objects.filter(**obj_fields).values_list('id', flat=True))

            many_to_many_objs[relatedmanager] = []
            val: BaseModel
            for val in value:
                sub_obj = get_subobj(val)
                existing_object_ids.discard(sub_obj.id)
                subobjects.append(sub_obj)
                sub_transfer = transfer_to_orm(
                    val,
                    sub_obj,
                    exclude_unset=exclude_unset,
                    access=access,
                    action=action,
                    _just_return_objs=True,
                )
                subobjects += sub_transfer[0]
                objects_to_delete += sub_transfer[1]

                if is_direct_m2m:
                    many_to_many_objs[relatedmanager].append(sub_obj)

            if not is_direct_m2m:
                objects_to_delete += related_model.objects.filter(id__in=list(existing_object_ids))

        else:
            raise NotImplementedError

    if subobjects and not action:
        raise AssertionError('action is not defined but subobjects exist')

    if _just_return_objs:
        return subobjects, objects_to_delete

    if (
        action in (TransferAction.CREATE, TransferAction.SYNC, TransferAction.NO_SUBOBJECTS)
        and created_submodels is None
    ):
        with atomic():
            should_save: Callable[[models.Model], bool] = lambda obj: bool(
                not do_not_save_if_no_change
                or (isinstance(obj, DirtyFieldsMixin) and obj.get_dirty_fields(check_relationship=True))
            )
            if should_save(django_obj):
                django_obj.save()

            for manager, objs in many_to_many_objs.items():
                for obj in objs:
                    if should_save(obj):
                        obj.save()

                manager.set(objs)

            if action in (TransferAction.SYNC, TransferAction.NO_SUBOBJECTS):
                for obj in objects_to_delete:
                    obj.delete()

            if action in (TransferAction.CREATE, TransferAction.SYNC):
                for obj in subobjects:
                    if should_save(obj):
                        obj.save()


async def update_orm(
    model: Type[BaseModel], orm_obj: models.Model, input: BaseModel, *, access: Optional[Access] = None
) -> BaseModel:
    """
    Apply (partial) changes given in `input` to an orm_obj and return an instance of `model` with the full data of the orm including the updated fields.
    """
    warnings.warn("Use transfer_to_orm with exclude_unset=True instead of this function", category=DeprecationWarning)

    if access:
        check_field_access(input, access)

    data = await model.from_orm(orm_obj)
    input_dict: dict = input.dict(exclude_unset=True)

    def update(model: BaseModel, input: dict):
        for key, value in input.items():
            if isinstance(value, dict):
                attr = getattr(model, key)
                if attr is None:
                    setattr(model, key, model.__fields__[key].type_.parse_obj(value))

                else:
                    update(attr, value)

            else:
                setattr(model, key, value)

    update(data, input_dict)

    values, fields_set, validation_error = validate_model(model, data.dict())
    if validation_error:
        raise RequestValidationError(validation_error.raw_errors)

    transfer_to_orm(data, orm_obj)
    return data



from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union

from django.db import models
from django.db.models.fields import Field as DjangoField
from pydantic.fields import FieldInfo, Undefined, UndefinedType
from pydantic.typing import NoArgAnyCallable

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, MappingIntStrAny
    from typing_extensions import Self


class ORMFieldInfo(FieldInfo):
    __slots__ = FieldInfo.__slots__ + (
        'orm_field',
        'orm_method',
        'scopes',
        'is_critical',
        'sync_matching',
        'is_sync_matching_field',
    )

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.orm_field: Optional[Union[DjangoField, UndefinedType]] = kwargs.pop('orm_field', None)
        self.orm_method: Optional[Union[Callable[['Self'], Any], Callable[['Self', Any], None]]] = kwargs.pop(
            'orm_method', None
        )
        self.scopes: Optional[List[str]] = kwargs.pop('scopes', None)
        self.is_critical: bool = kwargs.pop('is_critical', False)
        self.sync_matching: Optional[List[Tuple[str, DjangoField]]] = kwargs.pop('sync_matching', None)
        self.is_sync_matching_field: bool = kwargs.pop('is_sync_matching_field', False)

        super().__init__(default, **kwargs)


def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    alias: str = None,
    title: str = None,
    description: str = None,
    exclude: Union['AbstractSetIntStr', 'MappingIntStrAny', Any] = None,
    include: Union['AbstractSetIntStr', 'MappingIntStrAny', Any] = None,
    const: bool = None,
    gt: float = None,
    ge: float = None,
    lt: float = None,
    le: float = None,
    multiple_of: float = None,
    max_digits: int = None,
    decimal_places: int = None,
    min_items: int = None,
    max_items: int = None,
    unique_items: bool = None,
    min_length: int = None,
    max_length: int = None,
    allow_mutation: bool = True,
    regex: str = None,
    discriminator: str = None,
    repr: bool = True,
    orm_field: Optional[Union[DjangoField, UndefinedType]] = None,
    orm_method: Optional[Union[Callable[['Self'], Any], Callable[['Self', Any], None]]] = None,
    scopes: Optional[List[str]] = None,
    is_critical: bool = False,
    sync_matching: Optional[List[Tuple[str, DjangoField]]] = None,
    is_sync_matching_field: bool = False,
    **extra: Any,
) -> Any:
    if orm_field and isinstance(orm_field, models.CharField):
        if not max_length:
            max_length = orm_field.max_length

    field_info = ORMFieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        exclude=exclude,
        include=include,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        discriminator=discriminator,
        repr=repr,
        orm_field=orm_field,
        orm_method=orm_method,
        scopes=scopes,
        is_critical=is_critical,
        sync_matching=sync_matching,
        is_sync_matching_field=is_sync_matching_field,
        **extra,
    )
    field_info._validate()
    return field_info



import logging
import typing
from functools import cache
from types import FunctionType
from typing import Any, Callable, ForwardRef, Optional, Type, Union

from pydantic import BaseModel
from pydantic import Field as Field
from pydantic import create_model
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, FieldInfo, ModelField, Undefined
from pydantic.typing import get_origin, is_union

from ..fields import Field as ORMField
from ..fields import ORMFieldInfo

_logger = logging.getLogger(__name__)

TypingGenericAlias = type(Any)

_recreated_models = {}
_optional_models = {}


def _new_field_from_model_field(field: ModelField, default: Any = Undefined, required: Optional[bool] = None):
    if default is not Undefined:
        default = field.default

    if required is None and field.required and (default is Undefined or field.default is None):
        default = ...

    if isinstance(field.field_info, ORMFieldInfo):
        return ORMField(
            default,
            default_factory=field.default_factory,
            alias=field.alias,
            orm_field=field.field_info.orm_field,
            orm_method=field.field_info.orm_method,
            scopes=field.field_info.scopes,
            is_critical=field.field_info.is_critical,
            sync_matching=field.field_info.sync_matching,
            is_sync_matching_field=field.field_info.is_sync_matching_field,
            **field.field_info.extra,
        )

    return Field(
        default,
        default_factory=field.default_factory,
        alias=field.alias,
        **field.field_info.extra,
    )


class OptionalModel(BaseModel):
    pass


def optional_model(c, __module__: str, __parent__module__: str, id_key: str):
    try:
        if issubclass(c, BaseModel):
            if c in _optional_models:
                return _optional_models[c]

            field: ModelField
            fields = {}
            for key, field in c.__fields__.items():
                # TODO handle ForwardRef
                if field.shape == SHAPE_SINGLETON:
                    field_type = optional_model(
                        field.outer_type_,
                        __module__=__module__,
                        __parent__module__=__parent__module__,
                        id_key=id_key,
                    )

                else:
                    # TODO pydantic.get_origin ??
                    field_type = field.outer_type_

                default = field.default
                if key == id_key and not field.allow_none:
                    default = default or ...

                elif not field.allow_none:
                    field_type = Optional[field_type]

                elif field.required:
                    default = default or ...

                fields[key] = (field_type, _new_field_from_model_field(field, default, required=False))

            _logger.debug("Optional Model %s", c)
            _optional_models[c] = create_model(
                f'{c.__qualname__} [O]',
                __base__=(c, OptionalModel),
                __module__=c.__module__ if c.__module__ != __parent__module__ else __module__,
                **fields,
            )

            return _optional_models[c]

    except TypeError as error:
        pass

    return c


def to_optional(id_key: str = 'id'):
    def wrapped(cls: Type[BaseModel]):

        return optional_model(
            cls,
            __module__=cls.__module__,
            __parent__module__=cls.__base__.__module__,
            id_key=id_key,
        )

    return wrapped


class ReferencedModel(BaseModel):
    pass


class Reference(BaseModel):
    def __init_subclass__(cls, rel: Optional[str] = None, rel_params: Optional[Callable] = None, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls._rel = getattr(cls, '_rel', rel)
        cls._rel_params = rel_params
        c = cls
        while cls._rel is None:
            if issubclass(c.__base__, Reference):
                c = c.__base__
                if not c:
                    raise AssertionError("Cannot find parent Reference with `rel` set")

                try:
                    cls._rel = c._rel
                    cls._rel_params = c._rel_params

                except AttributeError:
                    pass

            else:
                raise AssertionError


def include_reference(reference_key: str = '$rel', reference_params_key: str = '$rel_params'):
    def wrapped(cls: Type[BaseModel]):
        def model_with_rel(c: Type, __parent__: Type, __module__: str, __parent__module__: str):
            if isinstance(c, ForwardRef):
                return c, False

            if is_union(get_origin(c)):
                models = [
                    model_with_rel(m, c, __module__=__module__, __parent__module__=__parent__module__)
                    for m in c.__args__
                ]
                return Union[(m[0] for m in models)], any(m[1] for m in models)

            if issubclass(c, BaseModel):
                field: ModelField
                fields = {}
                recreate_model = False
                for key, field in c.__fields__.items():
                    if field.shape not in (SHAPE_SINGLETON, SHAPE_LIST):
                        fields[key] = (field.outer_type_, _new_field_from_model_field(field))
                        continue

                    field_type, recreated_model = model_with_rel(
                        field.type_, c, __module__=__module__, __parent__module__=__parent__module__
                    )
                    if field.type_ != field.outer_type_:
                        field_type = getattr(typing, field.outer_type_._name)[field_type]

                    if field.allow_none:
                        field_type = Optional[field_type]

                    fields[key] = (field_type, _new_field_from_model_field(field))
                    if recreated_model:
                        recreate_model = True

                    try:
                        if issubclass(field_type, Reference):
                            recreate_model = True

                    except TypeError:
                        pass

                if issubclass(c, Reference):
                    recreate_model = True
                    value = Undefined
                    value_example = None
                    value_factory = None
                    if isinstance(c._rel, FunctionType):
                        value_factory = c._rel
                        value_example = c._rel()

                    else:
                        value = value_example = c._rel

                    fields['x_reference_key'] = (
                        str,
                        Field(
                            value,
                            example=value_example,
                            orm_field=None,
                            alias=reference_key,
                            default_factory=value_factory,
                        ),
                    )
                    if c._rel_params:
                        fields['x_reference_params_key'] = (
                            dict,
                            Field(alias=reference_params_key, orm_method=c._rel_params),
                        )

                if recreate_model:
                    if c not in _recreated_models:
                        _logger.debug(
                            "Recreate Model %s (in module %s)",
                            c,
                            c.__module__ if c.__module__ != __parent__module__ else __module__,
                        )
                        _recreated_models[c] = create_model(
                            f'{c.__qualname__} [R]',
                            __base__=(c, ReferencedModel),
                            __module__=c.__module__ if c.__module__ != __parent__module__ else __module__,
                            **fields,
                        )
                        _recreated_models[c].__recreated__ = True

                    if __parent__:
                        setattr(__parent__, c.__name__, _recreated_models[c])

                    return _recreated_models[c], True

            return c, False

        return model_with_rel(cls, None, __module__=cls.__module__, __parent__module__=cls.__base__.__module__)[0]

    return wrapped


def is_orm_field_set(field: FieldInfo) -> bool:
    if isinstance(field, ORMFieldInfo):
        orm_field = field.orm_field
        if orm_field is Undefined:
            return False

    else:
        orm_field = field.extra.get('orm_field')
        if 'orm_field' in field.extra and field.extra['orm_field'] is None:
            # Do not raise error when orm_field was explicitly set to None
            return False

    return True


def get_orm_field_attr(field: FieldInfo, key: str):
    if isinstance(field, ORMFieldInfo):
        return getattr(field, key)

    return field.extra.get(key)
